package jadex.adapter.jade.tools.logger;


import java.io.*;
import java.net.*;
import java.util.Map;
import java.util.*;
import java.util.logging.Level;
import jade.content.lang.Codec;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.*;
import jade.core.*;
import jade.core.behaviours.*;
import jade.domain.FIPAAgentManagement.*;
import jade.domain.*;
import jade.domain.introspection.*;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.*;
import jade.proto.SimpleAchieveREInitiator;
import jade.tools.ToolAgent;
import jade.util.ExpandedProperties;
import jade.util.leap.ArrayList;
import jade.util.leap.HashMap;
import jade.util.leap.Iterator;
import jade.util.leap.List;
import jade.util.leap.Set;
import jadex.adapter.jade.tools.logger.gui.MainWindow;
import jadex.adapter.jade.tools.logger.ontology.*;


/**
 * This is the <em>Logger</em> agent.
 * <br>
 * This class implements the low level part of the Logger, interacting with Jade
 * environment and with the logger GUI.<br>
 * At startup, the logger subscribes itself as an rma to be informed every time
 * an agent is born or dead, a container is created or deleted.
 * <p>
 * A properties file while may be used to control different logger properties.
 * These optional properties are as follows:
 * <ul>
 * <li>preload - A list of preload descriptions seperated by a semi-colon. Each description
 * consists of an agent name match string.
 * For details on the agent name match string, see the method isMatch().
 * If there is no @ in the agent name, it assumes the current HAP for it.
 *
 * <br>
 * Examples:
 * <pre>
 * preload=da0;da1
 * preload=agent??
 * preload=*
 * </pre>
 */
public class Logger extends ToolAgent
{

    private final Codec SL0 = new SLCodec(0);

    public final static Level defaultLevel = Level.FINE;

    public static final boolean LOG_ON = true;
    public static final boolean LOG_OFF = false;

    private java.util.Map preload = new java.util.HashMap();
    private ExpandedProperties properties = new ExpandedProperties();

    protected HashMap observedAgents = new HashMap();


    protected ArrayList agentsUnderLog = new ArrayList();
    protected ArrayList notifiers = new ArrayList();


    /*
     * Sends requests to the AMS for platform dependent actions
     * e.g. create LoggerNotifier
     */
    private class AMSClientBehaviour extends SimpleAchieveREInitiator
    {

        private String actionName;

        public AMSClientBehaviour(String an, ACLMessage request)
        {
            super(Logger.this, request);
            actionName = an;
        }

        protected void handleNotUnderstood(ACLMessage reply)
        {
            myGUI.showError("NOT-UNDERSTOOD received during " + actionName);
        }

        protected void handleRefuse(ACLMessage reply)
        {
            myGUI.showError("REFUSE received during " + actionName);
        }

        protected void handleAgree(ACLMessage reply)
        {
            // System.out.println("AGREE received");
        }

        protected void handleFailure(ACLMessage reply)
        {
            myGUI.showError("FAILURE received during " + actionName);
        }

        protected void handleInform(ACLMessage reply)
        {
            // System.out.println("INFORM received");
        }

    } // End of AMSClientBehaviour class


    private class ControlListenerBehaviour extends CyclicBehaviour
    {

        private MessageTemplate listenControlTemplate;

        ControlListenerBehaviour()
        {
            listenControlTemplate = MessageTemplate.MatchConversationId(getName() + "-control");
        }

        public void action()
        {

            ACLMessage current = receive(listenControlTemplate);
            if (current != null)
            {
                try
                {
                    Done d = (Done) getContentManager().extractContent(current);
                    Action a = (Action) d.getAction();
                    AID notifierID = a.getActor();

                    if (a.getAction() instanceof StartNotify)
                    {
                        StartNotify sn = (StartNotify) a.getAction();
                        AID agentID = sn.getObserved();
                        // add agent to HashMap
                        observedAgents.put(agentID, notifierID);
                        // add agent to  agentlist
                        myGUI.mainPanel.listAgent.addAgent(agentID);
                    }

                    if (a.getAction() instanceof StopNotify)
                    {
                        StopNotify sn = (StopNotify) a.getAction();
                        AID agentID = sn.getObserved();
                        // remove agent from HasMap
                        observedAgents.remove(agentID);
                        // remove agent from agentlist
                        myGUI.mainPanel.listAgent.removeAgent(agentID);
                    }
                }
                catch (Throwable e)
                {
                    //System.out.println("Serious problem Occurred");
                    myGUI.showError("An error occurred parsing the incoming message.\n" + "          The message was lost.");
                    e.printStackTrace();
                }
            }
            else
                block();
        }

    } // End of ControlListenerBehaviour


    private class LogListenerBehaviour extends CyclicBehaviour
    {

        private MessageTemplate listenLogTemplate;

        LogListenerBehaviour()
        {
            listenLogTemplate = MessageTemplate.MatchConversationId(getName() + "-event");
        }

        public void action()
        {

            ACLMessage current = receive(listenLogTemplate);
            if (current != null)
            {
                try
                {
                    Occurred o = (Occurred) getContentManager().extractContent(current);
                    EventRecord e = o.getWhat();
                    if (e.getWhat() instanceof PublishedLogRecord) {
                        PublishedLogRecord logData = (PublishedLogRecord) e.getWhat();
                        myGUI.mainPanel.logTable.addLogData(logData);
                    }

                }
                catch (Throwable e)
                {
                    //System.out.println("Serious problem Occurred");
                    myGUI.showError("An error occurred parsing the incoming message.\n" + "          The message was lost.");
                    e.printStackTrace();
                }
            }
            else
                block();
        }

    } // End of LogListenerBehaviour

    private class handleAddRemotePlatformBehaviour extends AMSClientBehaviour
    {

        public handleAddRemotePlatformBehaviour(String an, ACLMessage request)
        {
            super(an, request);

        }

        protected void handleInform(ACLMessage msg)
        {
            //System.out.println("arrived a new APDescription");
            try
            {
                AID sender = msg.getSender();
                Result r = (Result) getContentManager().extractContent(msg);

                Iterator i = r.getItems().iterator();
                APDescription APDesc = (APDescription) i.next();
                if (APDesc != null)
                {
                    myGUI.addRemotePlatformFolder();
                    myGUI.addRemotePlatform(sender, APDesc);
                }
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        }

    }//end handleAddRemotePlatformBehaviour

    private class handleRefreshRemoteAgentBehaviour extends AMSClientBehaviour
    {

        private APDescription platform;

        public handleRefreshRemoteAgentBehaviour(String an, ACLMessage request, APDescription ap)
        {
            super(an, request);
            platform = ap;

        }

        protected void handleInform(ACLMessage msg)
        {
            //System.out.println("arrived a new agents from a remote platform");
            try
            {
                AID sender = msg.getSender();
                Result r = (Result) getContentManager().extractContent(msg);
                Iterator i = r.getItems().iterator();
                myGUI.addRemoteAgentsToRemotePlatform(platform, i);
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        }

    }//end handleAddRemotePlatformBehaviour


    /**
     * Search keys in preload for a string which matches (using Pattern.matches method)
     * the agent name.
     * @param agentName The agent name.
     * @return String The key which matched.
     */
    protected String preloadContains(String agentName)
    {
        String preloadDescriptions = properties.getProperty("Preload", null);
        if (preloadDescriptions != null)
        {
            StringTokenizer parser = new StringTokenizer(preloadDescriptions, ";");
            while (parser.hasMoreElements())
            {
                String key = parsePreloadDescription(parser.nextToken());
                if (isMatch(key, agentName))
                {
                    return key;
                }
            }
        }
        return null;
    }

    /**
     * Converts a local name to the full name, if no expression is given.
     */
    private String parsePreloadDescription(String aDescription)
    {
        StringTokenizer st = new StringTokenizer(aDescription);
        String name = st.nextToken();
        if (!name.endsWith("*"))
        {
            int atPos = name.lastIndexOf('@');
            if (atPos == -1)
            {
                name = name + "@" + getHap();
            }
        }
        return name;
    }


    /**
     * Given two strings determine if they match. We iterate over the match expression
     * string from left to right as follows:
     * <ol>
     * <li> If we encounter a '*' in the expression token they match.
     * <li> If there aren't any more characters in the subject string token they don't match.
     * <li> If we encounter a '?' in the expression token we ignore the subject string's
     * character and move on to the next iteration.
     * <li> If the character in the expression token isn't equal to the character in
     * the subject string they don't match.
     * </ol>
     * If we complete the iteration they match only if there are the same number of
     * characters in both strings.
     * @param aMatchExpression An expression string with special significance to '?' and '*'.
     * @param aString The subject string.
     * @return True if they match, false otherwise.
     */
    protected boolean isMatch(String aMatchExpression, String aString)
    {
        int expressionLength = aMatchExpression.length();
        for (int i = 0; i < expressionLength; i++)
        {
            char expChar = aMatchExpression.charAt(i);
            if (expChar == '*')
                return true;   // * matches the remainder of anything
            if (i == aString.length())
                return false;  // if we run out of characters they don't match
            if (expChar == '?')
                continue;      // ? matches any single character so keep going
            if (expChar != aString.charAt(i))
                return false;  // if non wild then must be exactly equal
        }
        return (expressionLength == aString.length());
    }

    private SequentialBehaviour AMSSubscribe = new SequentialBehaviour();

    /**
     @serial
     */
    protected MainWindow myGUI;

    /**
     @serial
     */
    private String myContainerName;

    class LoggerAMSListenerBehaviour extends AMSListenerBehaviour
    {

        protected void installHandlers(Map handlersTable)
        {

            // Fill the event handler table.

            /* since snapshot 030922
            handlersTable.put(IntrospectionVocabulary.META_RESETEVENTS, new EventHandler()
            {
                public void handle(Event ev)
                {
                    ResetEvents re = (ResetEvents) ev;
                    myGUI.resetTree();
                }
            });
            */

            handlersTable.put(IntrospectionVocabulary.ADDEDCONTAINER, new EventHandler()
            {
                public void handle(Event ev)
                {
                    AddedContainer ac = (AddedContainer) ev;
                    ContainerID cid = ac.getContainer();
                    String name = cid.getName();
                    String address = cid.getAddress();
                    try
                    {
                        InetAddress addr = InetAddress.getByName(address);
                        myGUI.addContainer(name, addr);
                    }
                    catch (UnknownHostException uhe)
                    {
                        myGUI.addContainer(name, null);
                    }
                }
            });

            handlersTable.put(IntrospectionVocabulary.REMOVEDCONTAINER, new EventHandler()
            {
                public void handle(Event ev)
                {
                    RemovedContainer rc = (RemovedContainer) ev;
                    ContainerID cid = rc.getContainer();
                    String name = cid.getName();
                    myGUI.removeContainer(name);
                }
            });

            handlersTable.put(IntrospectionVocabulary.BORNAGENT, new EventHandler()
            {
                public void handle(Event ev)
                {
                    BornAgent ba = (BornAgent) ev;
                    ContainerID cid = ba.getWhere();
                    String container = cid.getName();
                    AID agent = ba.getAgent();
                    myGUI.addAgent(container, agent);
                    if (agent.equals(getAID()))
                        myContainerName = container;
                    // Here we check to see if the agent is one that we automatically will
                    // start logging.  If so, we invoke start the logging process.
                    if (preloadContains(agent.getName()) != null)
                    {
                        // System.out.println("Found one: " + agent.getName());
                        ArrayList agentList = new ArrayList();
                        agentList.add(agent);
                        logMsg(agentList, cid, getAMS(), true);
                        agentList.clear();
                    }
                    else
                    {
                        // System.out.println("Agent not in .inf: " + agent.getName());
                    }
                }
            });

            handlersTable.put(IntrospectionVocabulary.DEADAGENT, new EventHandler()
            {
                public void handle(Event ev)
                {
                    DeadAgent da = (DeadAgent) ev;
                    ContainerID cid = da.getWhere();
                    String container = cid.getName();
                    AID agent = da.getAgent();
                    myGUI.removeAgent(container, agent);

                }
            });

            handlersTable.put(IntrospectionVocabulary.MOVEDAGENT, new EventHandler()
            {
                public void handle(Event ev)
                {
                    MovedAgent ma = (MovedAgent) ev;
                    AID agent = ma.getAgent();
                    ContainerID from = ma.getFrom();
                    myGUI.removeAgent(from.getName(), agent);
                    ContainerID to = ma.getTo();
                    myGUI.addAgent(to.getName(), agent);
                }
            });

        }
    }	// END of inner class LoggerAMSListenerBehaviour

    /**
     * ACLMessages for subscription and unsubscription as <em>rma</em> are created and
     * corresponding behaviours are set up.
     */
    public void toolSetup()
    {

        // register ontology and language
        //        getContentManager().registerOntology(FIPAManagementOntology.getInstance());
        getContentManager().registerOntology(LoggerOntology.getInstance());
        //        getContentManager().registerLanguage(SL0);

        /**
         * Load properties from logger.properties file, if file exists
         */
        String fileName = properties.locatePropertyFile("logger.properties", null);
        //			properties.load(new FileInputStream("rma.properties"));
        //            String fileName = locateFile("logger.properties");
        if (fileName != null)
        {
            try
            {
                properties.addFromFile(fileName);
            }
            catch (IOException ioe)
            {
                // ignore - Properties not processed
            }
        }

        /*
         * preload agents from argument array if arguments present.
         * Note: this overrides the preload list from logger.properties file.
         */
        Object[] arguments = getArguments();
        if (arguments != null && arguments.length > 0)
        {
            String s = "";
            for (int i = 0; i < arguments.length; ++i)
            {
                s += arguments[i].toString();
            }
            properties.setProperty("Preload", s);

        }

        // Send 'subscribe' message to the client agent
        AMSSubscribe.addSubBehaviour(new SenderBehaviour(this, getSubscribe()));

        // Handle incoming 'inform' messages
        AMSSubscribe.addSubBehaviour(new LoggerAMSListenerBehaviour());

        // Schedule Behaviours for execution
        addBehaviour(AMSSubscribe);
        addBehaviour(new LogListenerBehaviour());
        addBehaviour(new ControlListenerBehaviour());

        // Show Graphical User Interface
        myGUI = new MainWindow(this, properties);
        myGUI.ShowCorrect();


    }

    private String locateFile(String aFileName)
    {
        try
        {
            String path = (new File(".")).getAbsolutePath();
            while (path != null)
            {
                path = path.replace('\\', '/');
                if (path.endsWith("."))
                {
                    path = path.substring(0, path.length() - 1);    // drop dot
                }

                if (path.endsWith("/"))
                {
                    path = path.substring(0, path.length() - 1);    // drop last separator
                }
                File dir = new File(path);
                File theFile = new File(dir, aFileName);
                if (theFile.exists())
                {
                    return theFile.getCanonicalPath();
                }
                else
                {
                    path = dir.getParent();    // reduce the path by one
                }
            }
        }
        catch (Exception any)
        {
            // ignore
        }
        return null;
    }

    /**
     * Cleanup during agent shutdown. This method cleans things up when
     * <em>Logger</em> agent is destroyed, disconnecting from client
     * agents and closing down the Logger administration <em>GUI</em>.
     * Currently logged agents are also unlogged to avoid errors.
     */
    protected void toolTakeDown()
    {
        Set keyset = observedAgents.keySet();
        Iterator iterator = keyset.iterator();
        while (iterator.hasNext())
        {
            AID agentID = (AID) iterator.next();
            AID notifier = (AID) observedAgents.get(agentID);

            // create ACLMessage for logging/unlogging agent
            ACLMessage requestMsg = new ACLMessage(ACLMessage.REQUEST);
            requestMsg.setSender(getAID());
            requestMsg.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
            requestMsg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            requestMsg.setOntology(LoggerOntology.ONTOLOGY_NAME);
            requestMsg.setConversationId(getName() + "-request");
            requestMsg.addReceiver(notifier);

            LogOff lo = new LogOff();
            lo.setLogger(getAID());
            lo.addLoggedAgent(agentID);

            Action a = new Action(getAID(), lo);

            try
            {
                getContentManager().fillContent(requestMsg, a);
            }
            catch (Codec.CodecException e)
            {
                e.printStackTrace();
            }
            catch (OntologyException e)
            {
                e.printStackTrace();
            }

            try
            {
                FIPAService.doFipaRequestClient(this, requestMsg, 1000);
            }
            catch (FIPAException e)
            {
                e.printStackTrace();
            }
        }


        // save properties to logger.properties
        /*
        try
        {
            Vector filter = myGUI.mainPanel.logTable.getFilter();
            for (int i = 0; i < filter.size(); i++)
            {
                Vector row = (Vector) filter.elementAt(i);
                StringBuffer sb = new StringBuffer();
                for (int j = 0; j < row.size(); j++)
                {
                    String cell = (String) row.elementAt(j);
                    sb.append(cell);
                    sb.append(";");
                }
                properties.setProperty("filter_" + i, sb.toString());
            }
            properties.setIntProperty("Width", myGUI.getSize().width);
            properties.setIntProperty("Height", myGUI.getSize().height);
            properties.setIntProperty("LocationX", myGUI.getLocation().x);
            properties.setIntProperty("LocationY", myGUI.getLocation().y);

            properties.setBooleanProperty("EnableFilter", myGUI.mainPanel.logTable.isDoFilter());
            properties.setBooleanProperty("ShowFilter", myGUI.mainPanel.isFilterVisible());
            OutputStream out = new FileOutputStream("logger.properties");
            properties.store(out, "Logger properties");
            out.close();
        }
        catch (IOException e)
        {
            //            e.printStackTrace();
        }

        */

        //                myGUI.mainPanel.logTable.removeAllLogRecords();

        // Now we unsubscribe from the rma list
        send(getCancel());
        // myGUI.setVisible(false); Not needed. Can cause thread deadlock.
        if(myGUI!=null)
        	myGUI.disposeAsync();
    }


    /**
     * This method add an LoggerNotifierBehaviour to perform a request
     * to the LoggerNotifier for logging/unlogging agents.
     **/
    public void logMsg(List agents, ContainerID cid, AID amsID, boolean onFlag)
    {
        Iterator it = agents.iterator();
        while (it.hasNext())
        {
            AID agentID = (AID) it.next();
            AID notifierID = (AID) observedAgents.get(agentID);
            addBehaviour(new RequestLoggingBehaviour(this, agentID, cid, amsID, notifierID, onFlag));

            //            System.out.println(sb.isRunnable());
            //            sb.addSubBehaviour(new LoggerNotifierBehaviour(this, agentID, amsID, onFlag));
            //            if (sb.isRunnable())
            //            {
            //                addBehaviour(sb);
            //            }
        }

    }


    //this method sends a request to a remote AMS to know the APDescription of a remote Platform
    public void addRemotePlatform(AID remoteAMS)
    {
        //System.out.println("AddRemotePlatform"+remoteAMS.toString());
        try
        {

            ACLMessage requestMsg = new ACLMessage(ACLMessage.REQUEST);
            requestMsg.setSender(getAID());
            requestMsg.clearAllReceiver();
            requestMsg.addReceiver(remoteAMS);
            requestMsg.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
            requestMsg.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            requestMsg.setOntology(FIPAManagementOntology.NAME);

            GetDescription action = new GetDescription();
            List l = new ArrayList(1);
            Action a = new Action();
            a.setActor(remoteAMS);
            a.setAction(action);

            getContentManager().fillContent(requestMsg, a);
            addBehaviour(new handleAddRemotePlatformBehaviour("GetDescription", requestMsg));

        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    public void removeRemotePlatform(APDescription platform)
    {
        myGUI.removeRemotePlatform(platform.getName());
    }


    //make a search on a specified ams in order to return
    //all the agents registered with that ams.
    public void refreshRemoteAgent(APDescription platform, AID ams)
    {
        try
        {
            // FIXME. Move all this block into the constructor for better performance
            // because it is invariant to the method parameters
            ACLMessage request; // variable that keeps the request search message
            Action act;  // variable that keeps the search action
            request = new ACLMessage(ACLMessage.REQUEST);
            request.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
            request.setLanguage(FIPANames.ContentLanguage.FIPA_SL0);
            request.setOntology(FIPAManagementOntology.NAME);
            AMSAgentDescription amsd = new AMSAgentDescription();
            SearchConstraints constraints = new SearchConstraints();
            constraints.setMaxResults(new Long(-1)); // all results back
            // Build a AMS action object for the request
            Search s = new Search();
            s.setDescription(amsd);
            s.setConstraints(constraints);
            act = new Action();
            act.setAction(s);

            // request has been already initialized in the constructor
            request.clearAllReceiver();
            request.addReceiver(ams);
            act.setActor(ams); // set the actor of this search action
            getContentManager().fillContent(request, act);

            addBehaviour(new handleRefreshRemoteAgentBehaviour("search", request, platform));

        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    public String getPreload()
    {
        return properties.getProperty("Preload", null);
    }

    public void setPreload(String newPreload)
    {
        properties.setProperty("Preload", newPreload);
    }


}  // End of class Logger
